#include <iostream>

#include "cuda_runtime.h"
#include "device_launch_parameters.h"

#include <random>

#include "CudaWorker.cuh"
#include "Gpu.h"

using namespace std;

__device__ __host__ void do_pythagorean(const float* p_a, const float* p_b, float* p_result) {
	float a = *p_a;
	float b = *p_b;
	*p_result = sqrt(a * a + b * b);
}

__global__ void pythagorean_kernel(const float* a, const float* b, float* result) {
	const size_t index = threadIdx.x + blockIdx.x * blockDim.x;
	if (index < VECTOR_SIZE) {
		do_pythagorean(a+index, b+index, result+index);
	}
}

CudaWorker::CudaWorker() {
	// Preallocate the vector storage
	a.reserve(VECTOR_SIZE);
	b.reserve(VECTOR_SIZE);
	cpu_result.reserve(VECTOR_SIZE);
	gpu_result.reserve(VECTOR_SIZE);

	// Initialize a random number generator
	std::default_random_engine rgen(83903);
	std::uniform_real_distribution<float> fran(0.0, 100.0);

	// Fill the two vectors with some random numbers
	for (size_t i = 0; i < VECTOR_SIZE; ++i) {
		a.push_back(fran(rgen));
		b.push_back(fran(rgen));
	}

	// Setup the same values on the GPU
	device_a = a;
	device_b = b;
	device_result.reserve(VECTOR_SIZE);
}

CudaWorker::~CudaWorker() {

}

void CudaWorker::CpuCompute() {
	for (size_t i = 0; i < VECTOR_SIZE; ++i) {
		float* ptr_to_a = &a[i];
		float* ptr_to_b = &b[i];
		float* ptr_to_result = &cpu_result[i];
		do_pythagorean(ptr_to_a, ptr_to_b, ptr_to_result);
	}
}

void CudaWorker::GpuCompute() {
	const size_t num_threads = Gpu::MaxThreadsPerBlock();
	const size_t num_blocks = VECTOR_SIZE / num_threads + 1;	// Simple ceiling function
	pythagorean_kernel<<<num_blocks, num_threads>>>(device_a.data().get(), device_b.data().get(), device_result.data().get());

	cudaError_t cudaStatus = cudaGetLastError();
	if (cudaStatus != cudaSuccess) {
		cout << "CudaWorker::GpuCompute() kernel failed with" << cudaGetErrorString(cudaStatus) << endl;
		return;
	}
	cudaStatus = cudaDeviceSynchronize();
	if (cudaStatus != cudaSuccess) {
		cout << "CudaWorker::GpuCompute() synchronize failed with " << cudaGetErrorString(cudaStatus) << endl;
		return;
	}
}

bool CudaWorker::Verify() {
	// Copy the GPU / device results back to the host
	gpu_result = device_result;

	// Make sure the size and contents of the CPU results matches the GPU results
	bool verification_result = cpu_result.size() == gpu_result.size();
	verification_result &= std::equal(cpu_result.begin(), cpu_result.end(), gpu_result.begin());

	return verification_result;
}
